from .deploy import deploy
from .exec import exec
from .init import init
from .parse import parse
from .server_command import server_command
from .ssh import ssh

__all__ = ["deploy", "exec", "init", "ssh", "server_command", "parse"]



import re
from pathlib import Path

import tomli

from dtuhpc.jobwriter.commands import Command, LoadModuleCommand
from dtuhpc.jobwriter.job_writer import JobWriter
from dtuhpc.jobwriter.options import (
    CoreBlockSizeOption,
    CorePTileSizeOption,
    EmailForNotificationsOption,
    ErrorOutputFilePathOption,
    MemoryPerCoreKillLimitOption,
    MemoryPerCoreOption,
    NameOption,
    NCPUCoresOption,
    Option,
    QueueOption,
    SendNotificationAtEndOption,
    SendNotificationAtStartOption,
    SingleHostOption,
    StandardOutputFilePathOption,
    UseGPUOption,
    WallTimeOption,
)

MAPPING = {
    "core_block_size": CoreBlockSizeOption,
    "core_p_tile_size": CorePTileSizeOption,
    "email": EmailForNotificationsOption,
    "error_output": ErrorOutputFilePathOption,
    "memory_kill_limit": MemoryPerCoreKillLimitOption,
    "memory": MemoryPerCoreOption,
    "cpu": NCPUCoresOption,
    "name": NameOption,
    "queue": QueueOption,
    "single_host": SingleHostOption,
    "standard_output": StandardOutputFilePathOption,
    "notification_start": SendNotificationAtStartOption,
    "notification_end": SendNotificationAtEndOption,
    "walltime": WallTimeOption,
    "use_gpu": UseGPUOption,
    "option": Option,
    "module": LoadModuleCommand,
    "commands": Command,
}


class JobReader:
    job_file_path: Path
    config: dict
    job_writer: JobWriter

    def __init__(self, job_file_path: str, variables: dict = None):
        self.job_file_path = Path(job_file_path)
        self.job_writer = JobWriter()
        self.load_job(job_file_path, variables)

    def load_job(self, job_file_path: str, variables: dict = None) -> "JobReader":
        with open(job_file_path, "r") as f:
            contents = self.replace_variables(f.read(), variables or {})
            self.config = tomli.loads(contents)

        return self

    @staticmethod
    def replace_variables(string: str, variables: dict) -> str:
        regex = re.compile(r"\${{\s*(?P<var_name>[a-zA-Z_]+[0-9]*)\s*}}")

        for match in regex.finditer(string):
            var_name = match.group("var_name")
            if var_name in variables:
                string = string.replace(match.group(0), variables[var_name])
            else:
                raise ValueError(f"Variable {var_name} not found in job file")

        return string

    def parse(self) -> "JobReader":
        for key, value in self.config.items():
            if key in MAPPING:
                option = MAPPING[key]

                if type(value) is dict:
                    self.job_writer.add(option(**value))
                elif type(value) is list:
                    for item in value:
                        self.job_writer.add(
                            option(**item) if type(item) is dict else option(item)
                        )
                else:
                    self.job_writer.add(option(value))
            else:
                raise ValueError(f"Unknown option: {key}")
        return self

    def to_str(self):
        return self.job_writer.to_string()



from typing import Optional

from .command import Command


class Showstart(Command):
    base_command = "showstart"

    def run(
        self,
        *job_ids: list[str],
        user: Optional[str] = None,
        queue: Optional[str] = None,
    ):
        args = []

        if user is not None:
            args.append(f"-u {user}")
        if queue is not None:
            args.append(f"-q {queue}")

        args += job_ids

        return super().run(*args)



import os
from io import StringIO
from typing import Optional

import click

from dtuhpc.cli.cli_config import CLIConfig
from dtuhpc.console import console
from dtuhpc.jobwriter.job_reader import JobReader


@click.command()
@click.option("--pr", "-p", default=None, type=int)
@click.option("--branch", "-b", default=None, type=str)
@click.argument("job_name", type=str, default=None)
@click.pass_obj
def deploy(
    config: CLIConfig, pr: Optional[int], branch: Optional[str], job_name: Optional[str]
):
    config.load_config()
    """Deploy a job."""
    gh = config.github()

    repo = config.git_repo()
    repo_remote = repo.remote("origin")
    repo_url = repo_remote.url
    repo.close()

    repo_id = (
        repo_url.replace("git@github.com:", "")
        .replace("https://github.com/", "")
        .replace(".git", "")
    )

    gh_repo = gh.get_repo(repo_id)

    if pr is not None:
        pr = gh_repo.get_pull(pr)
        branch_name = pr.head.ref
    elif branch is not None:
        branch_name = branch
    else:
        method = console.prompt_list("Branch or PR: ", ["Branch", "PR"])

        if method == 0:
            all_branches = gh_repo.get_branches()
            all_branch_names = [branch.name for branch in all_branches]

            branch_index = console.prompt_list("Pick a branch: ", all_branch_names)
            branch_name = all_branch_names[branch_index]
        else:
            pull_requests = gh_repo.get_pulls(state="open", sort="created")

            options = [
                f"#{pr.number}: {pr.title} ({pr.head.ref})" for pr in pull_requests
            ]

            if len(options) == 0:
                console.error("No open pull requests.")
                os.sys.exit(1)

            option_idx = console.prompt_list("Pick a PR: ", options)
            pr = pull_requests[option_idx]
            branch_name = pr.head.ref

    config.cwd = config.config["project"]["path"]

    conn = config.connection()
    conn.run("git fetch")
    conn.run(f"git checkout {branch_name}")
    conn.run("git pull")

    job_reader = JobReader(job_name)
    job_reader.parse()
    job_contents = job_reader.to_str()

    deploy_job_path = os.path.join(
        config.config["ssh"]["default_cwd"], ".dtuhpc/", "deploy_job.sh"
    )
    conn.conn.put(StringIO(job_contents), deploy_job_path)

    conn.run(f"bsub -cwd {config.cwd} < {deploy_job_path}")
    conn.run(f"rm {deploy_job_path}")

    conn.close()



from .command import Command


class LoadModuleCommand(Command):
    module: str

    def __init__(self, module: str):
        self.module = module

        super().__init__(command=f"module load {module}")



"""This module contains the JobWriter class, which is used to write LSF job scripts."""
from sys import stdout
from typing import Union

from .commands import Command
from .options import Option


class JobWriter:
    """This class is used to write LSF job scripts.

    Attributes:
        options (list[Option]): List of options to add to the job script.
        commands (list[Command]): List of commands to add to the job script.
    """

    options: list[Option] = []
    commands: list[Command] = []

    def __init__(self, opts: list[Option] = [], cmds: list[Command] = []):
        """Initialize the JobWriter with the given options and commands.

        Args:
            opts (list[Option], optional): List of options to add to the job script.
            Defaults to [].
            cmds (list[Command], optional): List of commands to add to the job script.
            Defaults to [].
        """
        self.options = opts
        self.commands = cmds

    def add(self, opt_or_cmd: Union[Option, Command]):
        if isinstance(opt_or_cmd, Option):
            self.add_option(opt_or_cmd)
        if isinstance(opt_or_cmd, Command):
            self.add_command(opt_or_cmd)

        return self

    def add_option(self, option: Option) -> "JobWriter":
        """Add an option to the job script.

        Args:
            option (Option): Option to add to the job script.

        Returns:
            JobWriter: The JobWriter object.
        """
        self.options.append(option)
        return self

    def add_command(self, command: Command) -> "JobWriter":
        """Add a command to the job script.

        Args:
            command (Command): Command to add to the job script.

        Returns:
            JobWriter: The JobWriter object.
        """
        self.commands.append(command)
        return self

    def to_string(self) -> str:
        """Convert the job script to a string.

        Returns:
            str: The job script as a string.
        """

        lines = ["#!/bin/sh", "# LSF Job options"]
        lines += list(map(lambda o: o.get_option_line(), self.options))
        lines.append("# Commands")
        lines += list(map(lambda c: c.get_command(), self.commands))

        return "\n".join(lines)

    def to_stdout(self):
        """Write the job script to stdout.

        Returns:
            JobWriter: The JobWriter object.
        """
        stdout.write(self.to_string())
        return self



from sys import stdout

import click

from dtuhpc.jobwriter.job_reader import JobReader


@click.command()
@click.argument("job_path", nargs=1, type=str, required=True)
def parse(job_path: str):
    """Initiates the current project on DTU's HPC server."""

    job_reader = JobReader(job_path)
    job_reader.parse()

    stdout.write(job_reader.to_str())
    stdout.close()



from .option import Option


class MemoryPerCoreKillLimitOption(Option):
    memory_in_gb_per_core: int

    def __init__(self, memory_per_core_in_gb: int):
        self.memory_in_gb_per_core = memory_per_core_in_gb
        super().__init__(option="M", value=self.format_ram_option())

    def format_ram_option(self):
        return f"rusage[mem={self.memory_in_gb_per_core}GB]"



import os
from io import StringIO

import click

from dtuhpc.cli.cli_config import CLIConfig
from dtuhpc.commands import BSub
from dtuhpc.jobwriter.job_reader import JobReader


@click.command()
@click.pass_obj
def init(config: CLIConfig):
    """Initiates the current project on DTU's HPC server."""
    config.load_config()
    repo = config.git_repo()
    repo_remote = repo.remote("origin")
    repo_url = repo_remote.url
    repo.close()

    project_name = config.config["project"]["name"]
    project_path = config.config["project"]["path"]
    variables = {
        "project_name": project_name,
        "project_path": project_path,
        "git_url": repo_url,
    }

    job_reader = JobReader("default_jobs/init-poetry.toml", variables)
    job_reader.parse()

    conn = config.connection()

    conn.run("mkdir -p .dtuhpc")
    conn.conn.put(
        StringIO(job_reader.to_str()),
        os.path.join(config.cwd, ".dtuhpc/", f"initialize_{project_name}_job.sh"),
    )

    bsub = BSub(conn)
    bsub.run(f"< .dtuhpc/initialize_{project_name}_job.sh")

    conn.close()



from .core_block_size_option import CoreBlockSizeOption
from .core_ptile_size_option import CorePTileSizeOption
from .email_for_notifications_option import EmailForNotificationsOption
from .error_output_file_path_option import ErrorOutputFilePathOption
from .memory_per_core_kill_limit_option import MemoryPerCoreKillLimitOption
from .memory_per_core_option import MemoryPerCoreOption
from .n_cpu_cores_option import NCPUCoresOption
from .name_option import NameOption
from .option import Option
from .queue_option import QueueOption
from .send_notification_at_end_option import SendNotificationAtEndOption
from .send_notification_at_start_option import SendNotificationAtStartOption
from .single_host_option import SingleHostOption
from .standard_output_file_path_option import StandardOutputFilePathOption
from .use_gpu_option import UseGPUOption
from .wall_time_option import WallTimeOption

__all__ = [
    "CoreBlockSizeOption",
    "CorePTileSizeOption",
    "EmailForNotificationsOption",
    "ErrorOutputFilePathOption",
    "MemoryPerCoreKillLimitOption",
    "MemoryPerCoreOption",
    "NCPUCoresOption",
    "NameOption",
    "Option",
    "QueueOption",
    "SendNotificationAtStartOption",
    "SendNotificationAtEndOption",
    "SingleHostOption",
    "StandardOutputFilePathOption",
    "SingleHostOption",
    "UseGPUOption",
    "WallTimeOption",
]



from .option import Option


class WallTimeOption(Option):
    hours: int
    minutes: int

    def __init__(self, hours: int = 0, minutes: int = 0):
        assert hours >= 0 and minutes >= 0
        assert hours > 0 or minutes > 0

        self.hours = hours
        self.minutes = minutes

        super().__init__(option="W", value=self.format_time())

    @staticmethod
    def time_prefix(time: int) -> str:
        return f"0{time}" if time < 10 else str(time)

    def format_time(self):
        return f"{self.time_prefix(self.hours)}:{self.time_prefix(self.minutes)}"



from .option import Option


class CorePTileSizeOption(Option):
    ptile_size: int

    def __init__(self, ptile_size: int):
        assert ptile_size > 0
        self.ptile_size = ptile_size

        super().__init__(option="R", value=f"span[ptile={ptile_size}]")



import base64
import json
import os
from pathlib import Path
from typing import Optional

import click
import tomli
from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from git import Repo
from github import Github

from dtuhpc.connector import HPCConnection
from dtuhpc.console import console


class CLIConfig:
    config_path: Path
    hide: bool
    config: Optional[dict] = None
    auth: dict = {}
    cwd: Optional[str]

    def __init__(
        self,
        config_path: Optional[str] = None,
        hide: bool = True,
        cwd: Optional[str] = None,
    ):
        self.config_path = (
            Path(config_path) if config_path is not None else self._get_config_path()
        )
        self.hide = hide
        self.cwd = cwd

    @staticmethod
    def _get_global_config_path() -> Path:
        return Path(os.path.join(Path.home(), ".config", ".dtuhpc.toml"))

    @staticmethod
    def get_global_auth_path() -> Path:
        return Path(os.path.join(Path.home(), ".config", "dtuhpc", "auth.json"))

    @staticmethod
    def _get_local_config_path() -> Path:
        return Path(os.path.join(Path.cwd(), ".dtuhpc.toml"))

    def _get_config_path(self) -> Path:
        local_path = self._get_local_config_path()
        if local_path.exists():
            return local_path

        global_path = self._get_global_config_path()
        if global_path.exists():
            return global_path

        console.error("Could not find any config file.")
        os.sys.exit(1)

    def load_config(self):
        with open(self.config_path, "rb") as config_file:
            config = tomli.load(config_file)

        self.config = config
        self.cwd = self.cwd if self.cwd is not None else config["ssh"]["default_cwd"]

    def _load_auth(self) -> dict:
        if not os.path.exists(self.get_global_auth_path()):
            console.error("Could not find auth file. Please call 'dtuhpc auth' first.")
            os.sys.exit(1)

        with open(self.get_global_auth_path(), "rb") as auth_file:
            auth = json.load(auth_file)

        encryption_key = bytes(click.prompt("Encryption key", hide_input=True), "utf-8")

        salt = base64.urlsafe_b64decode(auth["salt"].encode("utf-8"))
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=480000,
        )
        key = base64.urlsafe_b64encode(kdf.derive(encryption_key))

        f = Fernet(key)

        auth["password"] = f.decrypt(bytes(auth["password"], "utf-8")).decode("utf-8")

        return auth

    def connection(self) -> HPCConnection:
        self.auth = self._load_auth()
        console.primary("Connecting...")
        conn = HPCConnection(
            user=self.auth["username"],
            host=self.config["ssh"]["host"],
            password=self.auth["password"],
            hide=self.hide,
            cwd=self.cwd,
        )
        console.success("Connected!")
        return conn

    def git_repo(self) -> Repo:
        return Repo(os.getcwd())

    def github(self) -> Github:
        return Github(self.config["github"]["access_token"])



from .bkill import bkill
from .bqueues import bqueues
from .bstat import bstat
from .bsub import bsub
from .nodestat import nodestat
from .showstart import showstart

__all__ = ["bkill", "bqueues", "bstat", "bsub", "nodestat", "showstart"]



import click

from dtuhpc.cli.cli_config import CLIConfig
from dtuhpc.cli.commands import deploy, exec, init, parse, server_command, ssh
from dtuhpc.cli.commands.auth import auth


@click.group()
@click.option(
    "--hide/--no-hide",
    default=False,
    show_default=True,
    help="hide output from the server.",
)
@click.option(
    "--config", default=None, type=click.Path(exists=True), help="path to config file."
)
@click.option(
    "--cwd",
    default=None,
    type=str,
    help="default working directory for executing commands.",
)
@click.pass_context
def cli(ctx, hide, config, cwd):
    ctx.obj = CLIConfig(config_path=config, hide=hide, cwd=cwd)


def main():
    cli.add_command(server_command)
    cli.add_command(init)
    cli.add_command(exec)
    cli.add_command(deploy)
    cli.add_command(ssh)
    cli.add_command(auth)
    cli.add_command(parse)

    cli()


if __name__ == "__main__":
    main()



import click

from dtuhpc.cli.cli_config import CLIConfig


@click.command()
@click.argument("command", nargs=1, type=str, required=True)
@click.pass_obj
def exec(config: CLIConfig, command: str):
    config.load_config()
    """Executes COMMAND on the DTU HPC Server."""
    conn = config.connection()
    conn.run(command)

    conn.close()
